
!***************************************************************************************************************************
!                                                                                                                          !
!  PROGRAM: BF-SEM                                                                                                            !
!  This program computes Bayes factors and posterior model probabilities for a set of inquality constrained hypotheses.    !
!  I provide the code such that everyone can verify for themselves.                                                        !
!  If you have any problem or find any bug of this program, please contact Xin Gu via email x.gu@uu.nl                     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                                                                                          !
!***************************************************************************************************************************

  include 'link_f90_static.h'   ! Compiler V9
! include 'link_fnl_static.h'   ! Compiler V10 		

program BFSEM

    USE IMSL_LIBRARIES
    
    ! Declare variables
    implicit none
    ! Integers
    integer                             :: h, i, j, k, ISEED
    integer								:: ios, numH, numP, numR, rowrank, error
    integer, dimension(:), allocatable  :: numf, numc1, numc2
    ! Reals/doubles
    real                                :: totalf, totalc1, totalc2, temp
    real, dimension(:), allocatable     :: theta, bet, betprior, ff, cc1, cc2, bf1, bf2, bfc1, bfc2, PMP1, PMP2
    real, dimension(:), allocatable     :: constant, transcon
    real, dimension(:,:), allocatable   :: Rr, thetacov, betacov, betaprior1, betaprior2, tempmatrix
    real, dimension(:,:), allocatable   :: invbetacov, invbetaprior1, invbetaprior2, invDiag, invDiagprior1, invDiagprior2
    real, dimension(:,:), allocatable   :: IE, B, priorB1, priorB2, transR
    character*256                       :: annotate

error=0 ! Record whether the input file is valid

! Open two .txt files: Input.txt and Output.txt
Open(unit = 10, file = 'Input.txt', status ='OLD', iostat=ios)
Open(unit = 20, file = 'Output.txt', status = 'replace')

! Check whether Input.txt exists.
if (ios .NE. 0) then
    error=1 ! not exsit
    go to 999
end if

! Read annotate line
read(10,'(A1)') annotate 
if (annotate .NE. '#') then
    error=22 ! annotate line is missing
    go to 999
end if

! Read the number of parameters (numP) and the number of hypotheses (numH) under consideration
read(10,*,iostat=ios) numP, numH
if (ios .NE. 0 .OR. numP .LE. 0 .OR. numH .LE. 0) then
    error=2 ! an error in this line
    go to 999
end if

! Allocate the estiamtes of parameters (theta) and their covariance matrix (thetacov)
allocate (theta(numP+1)) 
allocate (thetacov(numP,numP))

! Allocate the Bayes factors of all the hypotheses and the Posterior Model Probabilities, under two prior distributions 
allocate (bf1(numH),bf2(numH),bfc1(numH),bfc2(numH),PMP1(numH),PMP2(numH))

! Read annotate line
read(10,'(A1)') annotate
if (annotate .NE. '#') then
    error=33 ! annotate line is missing
    go to 999
end if

! Read the estimates of parameters(theta). This can be otained in Lavaan.
read(10,*,iostat=ios) theta(1:numP)
theta(numP+1) = -1  ! We add an extra -1 in theta so that Rtheta>r becomes [R|r]theta>0, where [R|r] is an augmented matrix.
if (ios .NE. 0) then
    error=3 ! an error in this line
    go to 999
end if

! Read annotate line
read(10,'(A1)') annotate
if (annotate .NE. '#') then
    error=44 ! annotate line is missing
    go to 999
end if

! Read the covariance matrix of parameters (thetacov). This can be otained in Lavaan.
do i = 1,numP
   read(10,*,iostat=ios) thetacov(i,1:numP) 
if (ios .NE. 0) then
    error=4 ! an error in this line
    go to 999
end if
end do

ISEED=12345
call RNSET(ISEED)! Initialize a random seed for IMSL random number generators


do h = 1,numH ! numH is the number of hypotheses under consideration. Each loop h is for one hypothesis 
   ! Read annotate line
   read(10,'(A1)',iostat=ios) annotate 
   ! Check whether all hypotheses are specified in Input.txt.
   if (ios .NE. 0) then
       error=5 ! One or more hypotheses are not specified in Input.txt
       go to 999
   end if
   ! Check whether annotate line is missing
   if (annotate .NE. '#') then
       error=66 ! annotate line is missing
       go to 999
   end if
     
   ! Read the number of constraints in the hypothesis h, h=1,...,numH. In the paper, we use i.  
   read(10,*,iostat=ios) numR
   if (ios .NE. 0 .OR. numR .LE. 0) then
       error=6 ! an error in this line
       go to 999
   end if

   ! Allocate the augmented restriction matrix [R|r] and the matrics A and C used in parameter transformation in Section 5.2
   allocate (Rr(numR,numP+1),constant(numR),transR(numR,numR),transcon(numR))

   ! Read annotate line
   read(10,'(A1)') annotate
   if (annotate .NE. '#') then
       error=77 ! annotate line is missing
       go to 999
   end if
   
   ! Read the augmented matrix [R|r]
   do i = 1, numR
      read(10,*,iostat=ios) Rr(i,1:(numP+1))
      if (ios .NE. 0) then
         error=7 ! an error in this line
         go to 999
      end if      
      constant(i)=Rr(i,numP+1)
   end do
          

   ! The following is the function for parameter transformation shown in the procedure in section 5.2, 
   ! which renders the row rank of R and the matrix C.
   call Mrank(numR, numP, rowrank, Rr, transR, constant, transcon)  ! Function is defined below in subroutine Mrank
   if(error == 8) then   ! The hypothesis input is not inequality constrained hypothesis.
   go to 999
   end if
   
   ! Allocate the vectors and matrics.
   allocate (ff(numR),cc1(numR),cc2(numR),numf(numR),numc1(numR),numc2(numR),bet(rowrank),betprior(rowrank))
   allocate (betacov(rowrank,rowrank),betaprior1(rowrank,rowrank),betaprior2(rowrank,rowrank),tempmatrix(rowrank,numP))
   allocate (invbetacov(rowrank,rowrank),invbetaprior1(rowrank,rowrank),invbetaprior2(rowrank,rowrank))
   allocate (invDiag(rowrank,rowrank),invDiagprior1(rowrank,rowrank),invDiagprior2(rowrank,rowrank))
   allocate (IE(rowrank,rowrank),B(rowrank,rowrank),priorB1(rowrank,rowrank),priorB2(rowrank,rowrank))  
         
   ! Parameter transformation: beta=R*theta-r, betacov=R*thetacov*R^(T)
   call MURRV(Rr(1:rowrank, 1:(numP+1)), theta(1:(numP+1)), bet)       ! beta=Rr*theta  
   call MXYTF(Rr(1:rowrank, 1:numP), Rr(1:rowrank,1:numP), betaprior1) ! betaprior1=R*IE*R^(T)=R*R^(T). betaprior1 is the covariance structure in prior (19)
   call MRRRR(Rr(1:rowrank, 1:numP), thetacov, tempmatrix)             ! tempmatrix=R*thetacov. 
   call MXYTF(tempmatrix, Rr(1:rowrank,1:numP), betacov)               ! betacov=tempmatrix*R^(T), then betacov=R*thetacov*R^(T).
                                                                       ! betacov is the covariance structure in prior (20) and the covariance in posterior (24).


      
   betaprior1=10000*betaprior1 ! let w=10000 in (19) 
   betaprior2=10000*betacov    ! let w=10000 in (20)

   ! Calculate the matrix used in the conditional distribution (39) for Gibbs sampling. (both prior and posterior)
   ! In equation (39), B=I-[diag(invbetacov)]^(-1)*invbetacov
   call LINRG(betacov,invbetacov)        ! Compute the inverse of betacov i.e., invbetacov=betacov^(-1)
   call LINRG(betaprior1,invbetaprior1)  ! Compute the inverse of betaprior1 i.e., invbetaprior1=betaprior1^(-1)
   call LINRG(betaprior2,invbetaprior2)  ! Compute the inverse of betaprior2 i.e., invbetaprior2=betaprior2^(-1)

   ! "invDiag" is the inverse matrix of the diagonal matrix of "invbetacov", i.e.,[diag(invbetacov)]^(-1)
      IE=0
   do i=1,rowrank
      invDiag(i,i)=(invbetacov(i,i))**(-1)
      invDiagprior1(i,i)=(invbetaprior1(i,i))**(-1)
      invDiagprior2(i,i)=(invbetaprior2(i,i))**(-1)
      IE(i,i)=1
   end do
    
   ! Compute B=I-[diag(invbetacov)]^(-1)*invbetacov, and also for two priors
   do i=1,rowrank
      do j=1,rowrank
         B(i,j)=IE(i,j)-invDiag(i,i)*invbetacov(i,j)
         priorB1(i,j)=IE(i,j)-invDiagprior1(i,i)*invbetaprior1(i,j)
         priorB2(i,j)=IE(i,j)-invDiagprior2(i,i)*invbetaprior2(i,j)
      end do
   end do  

   
   !Gibbs sampler procedure
    
   ff=0
   cc1=0
   cc2=0
   
   do k=1,numR   ! For each decomposed complexity and fit. 

      !  Function gibbsampler is defined below in subroutine gibbsampler. 
      call gibbsampler(k,rowrank,bet,transcon,B,invbetacov,transR,ff,numf)  ! fit, 
      print '("Constraint", I4 " for fit in hypothesis", I3  " has been evaluated", I10)', k, h
      
      betprior=0
      call gibbsampler(k,rowrank,betprior,transcon,priorB1,invbetaprior1,transR,cc1,numc1)  ! complexity under prior (19).
      print '("Constraint", I4 " for complexity (prior 1) in hypothesis", I3 " has been evaluated")' , k, h
      
      call gibbsampler(k,rowrank,betprior,transcon,priorB2,invbetaprior2,transR,cc2,numc2)  ! complexity under prior (20).
      print '("Constraint", I4 " for complexity (prior 2) in hypothesis", I3 " has been evaluated")' , k, h
   
   end do        

   ! Compute the complexity, fit and Bayes factor for hypothesis h under two priors
   totalf=1
   totalc1=1
   totalc2=1
   bf1(h)=1
   bf2(h)=1
   do k=1,numR
      totalf=totalf*ff(k)
      totalc1=totalc1*cc1(k)
      totalc2=totalc2*cc2(k)
      bf1(h)=bf1(h)*(ff(k)/cc1(k))
      bf2(h)=bf2(h)*(ff(k)/cc2(k))
   end do
   
   ! Compute the Bayes factor for inequality constrained hypothesis vs its compliment
   bfc1(h)=bf1(h)/((1-totalf)/(1-totalc1))
   bfc2(h)=bf2(h)/((1-totalf)/(1-totalc2))
   
! Output
write(20,'(A,I3)') 'Result for hypothesis', h
write(20,*)
write(20,'(A)') '     Fits                numbers of iterations'   
do k = 1, numR
   write(20,'(f10.4,I25)') ff(k), numf(k)
end do
write(20,*)
write(20,'(A)') 'Complexities (prior 1)   numbers of iterations' 
do k = 1, numR
   write(20,'(f10.4,I25)') cc1(k), numc1(k)
end do
write(20,*)
write(20,'(A)') 'Complexities (prior 2)   numbers of iterations' 
do k = 1, numR
   write(20,'(f10.4,I25)') cc2(k), numc2(k)
end do
write(20,*)
write(20,'(A)') 'Total fit      Complexity (prior1)     Complexity (prior 2)'
write(20,'(f8.4,f19.4,f23.4)') totalf, totalc1, totalc2
write(20,*)
write(20,'(A)') 'BFiu (prior 1)    BFiu (prior 2)'
write(20,'(f8.4,f19.4)') bf1(h), bf2(h)
write(20,*)
write(20,'(A)') 'BFic (prior 1)    BFic (prior 2)'
write(20,'(f8.4,f19.4)') bfc1(h), bfc2(h)
write(20,*)
write(20,*)
   
! deallocate variables used for the current hypothesis such that these can be used for the next hypothesis
deallocate(Rr,ff,cc1,cc2,numf,numc1,numc2,bet,betacov,betaprior1,betaprior2,tempmatrix,constant,transR,transcon,betprior)
deallocate(invbetacov,invbetaprior1,invbetaprior2,invDiag,invDiagprior1,invDiagprior2,IE,B,priorB1,priorB2)

end do

! Check whether there are extra lines that are not read.
read(10,'(A5)',iostat=ios) annotate
   if(ios == 0 .AND. annotate .NE. '') then
      error=9
      go to 999
   end if
close(10)

! report the PMPs for each hypothesis under two priors
write(20,*)
write(20,'(A)') 'Result of PMP for each hypothesis'
write(20,*)
do h = 1, numH
write(20,'(A,I3)') 'PMP (prior1)      PMP (prior2)    for hypothesis', h 
write(20,'(f8.4,f19.4)') bf1(h)/sum(bf1), bf2(h)/sum(bf2)
write(20,*)
write(20,*)
end do

! report error message in the case of that error is not equal to 0
999 write(*,*)
if(error == 1) then
   write (*,'(A)') 'WARNING: Input.txt does not exist'
   write (20,'(A)') 'WARNING: Input.txt does not exist'
   write(*,*)
   write(*,*)
   write (*,'(A)') 'Check the name of input file'
   write (20,'(A)') 'Check the name of input file'
else if (error == 2) then
   write(*,'(A)') 'WARNING: An error in the 2nd line of Input.txt'
   write(20,'(A)') 'WARNING: An error in the 2nd line of in Input.txt'
   write(*,*)
   write(20,*)
   write(*,'(A)') 'Two positive integers should be given here'
   write(20,'(A)') 'Two positive integers should be given here'
else if (error == 3) then
   write(*,'(A)') 'WARNING: An error below "#Estimates of parameters" in Input.txt'
   write(20,'(A)') 'WARNING: An error below "#Estimates of parameters" in Input.txt'
   write(*,*)
   write(20,*)
   write(*,'(A)') 'The length of estimates should be equal to the number of structural parameters'
   write(20,'(A)') 'The length of estimates should be equal to the number of structural parameters'
else if (error == 4) then
   write(*,'(A)') 'WARNING: An error below "#Covariance matrix of parameters" in Input.txt'
   write(20,'(A)') 'WARNING: An error below "#Covariance matrix of parameters" in Input.txt'
   write(*,*)
   write(20,*)
   write(*,'(A)') 'The rank of covariance matrix should be equal to the number of structural parameters'
   write(20,'(A)') 'The rank of covariance matrix should be equal to the number of structural parameters'  
else if (error == 5) then
   write(*,*)
   write(20,*)
   write(*,'(A,i3,A)') 'Hypothesis', h, '  needs to be specified in Input.txt'
   write(20,'(A,i3,A)') 'Hypothesis', h, '  needs to be specified in Input.txt'
else if (error == 6) then
   write(*,'(A,i3,A)') 'WARNING: An error below #Number of constraints in hypothesis', h, '  in Input.txt'
   write(20,'(A,i3,A)') 'WARNING: An error below #Number of constraints in hypothesis', h, '  in Input.txt'
   write(*,*)
   write(20,*)
   write(*,'(A,i3,A)') 'One positive number should be given'
   write(20,'(A,i3,A)') 'One positive number should be given'
else if (error == 7) then
   write(*,'(A,i3,A)') 'WARNING: An error below #Restriction matrix (R-r) for hypothesis', h, '  in Input.txt'
   write(20,'(A,i3,A)') 'WARNING: An error below #Restriction matrix (R-r) for hypothesis', h, '  in Input.txt'
   write(*,*)
   write(20,*)
   write(*,'(A,i3,A,i3,A)') 'There should be a', numR, ' *', numP+1, '  matrix' 
   write(20,'(A,i3,A,i3,A)') 'There should be a', numR, ' *', numP+1, '  matrix' 
else if (error == 8) then
   write(*,'(A,i3,A)') 'WARNING: Hypothesis', h, '  contains equality, about equality, range, or contradicting constraints! '
   write(20,'(A,i3,A)') 'WARNING: Hypothesis', h, '  contains equality, about equality, range, or contradicting constraints! ' 
else if (error == 9) then
   write(*,'(A)') 'WARNING: Not all of lines in Input.txt are read'   
   write(20,'(A)') 'WARNING: Not all of lines in Input.txt are read' 
   write(*,*)
   write(20,*)
   write(*,'(A)') 'Check the number of hypotheses in the 2nd line'
   write(20,'(A)') 'Check the number of hypotheses in the 2nd line'
   write(*,*)
   write(20,*)
   write(*,'(A,i3,A)') 'Or the number of rows in restriction matrix for hypothesis', numH, '  is not equal to the number of constraints'
   write(20,'(A,i3,A)') 'Or the number of rows in restriction matrix for hypothesis', numH, '  is not equal to the number of constraints'
else if (error == 22) then
   write(*,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(20,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(*,*)
   write(20,*)
   write(*,'(A)') '#Number of structural parameters and hypotheses under consideration'
   write(20,'(A)') '#Number of structural parameters and hypotheses under consideration'
else if (error == 33) then  
   write(*,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(20,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(*,*)
   write(20,*)
   write(*,'(A)') '#Estimates of parameters'
   write(20,'(A)') '#Estimates of parameters'
else if (error == 44) then
   write(*,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(20,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(*,*)
   write(20,*)
   write(*,'(A)') '#Covariance matrix of parameters'
   write(20,'(A)') '#Covariance matrix of parameters'
else if (error == 66) then
   write(*,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(20,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(*,*)
   write(20,*)
   write(*,'(A,i3)') '#Number of constraints in hypothesis', h
   write(20,'(A,i3)') '#Number of constraints in hypothesis', h
   if (h == 1) then
      write(*,*)
      write(20,*)
      write(*,'(A)') 'Or the number of rows in covariance matrix is not equal to the number of structural parameters'
      write(20,'(A)') 'Or the number of rows in covariance matrix is not equal to the number of structural parameters'
   else
      write(*,*)
      write(20,*)
      write(*,'(A,i3,A)') 'Or the number of rows in restriction matrix for hypothesis', h-1, '  is not equal to the number of constraints'
      write(20,'(A,i3,A)') 'Or the number of rows in restriction matrix for hypothesis', h-1, '  is not equal to the number of constraints'
   end if
else if (error == 77) then
   write(*,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(20,'(A)') 'WARNING: Miss an annotate line in Input.txt:'
   write(*,*)
   write(20,*)
   write(*,'(A,i3)') '#Restriction matrix (R-r) for hypothesis', h
   write(20,'(A,i3)') '#Restriction matrix (R-r) for hypothesis', h
end if
 
! The pause is used when an error occurs. 
if(error .NE. 0) then 
   write(*,*)
   write(*,*)
   pause 'Stop the program by pressing enter'         
end if

close(20)     


! Function for parameter transformation. (rowrank denotes M, transR denotes C in the paper)
! It uses a modified Gaussian elimination algorithm.
contains
subroutine Mrank(numR, numP, rowrank, Rr, transR, constant, transcon)
   
   ! define variables
   implicit none
   integer, intent(in)                                     :: numR, numP
   integer, intent(out)                                    :: rowrank
   Integer                                                 :: i, j, k
   real                                                    :: temp
   real, allocatable, dimension(:)                         :: rownumber
   real, dimension(:), intent(inout)                       :: constant, transcon
   real, allocatable, dimension(:,:)                       :: A, Ainv
   real, dimension(:,:), intent(inout)                     :: Rr, transR

   
   ! Allocate
   if (numR>numP) then    
      allocate (A(numR,numR), Ainv(numR,numR), rownumber(numR))
   else 
      allocate (A(numR,numP), Ainv(numR,numP), rownumber(numR))
   end if   
     
   A=0
   Ainv=0
   transR=0
   rownumber=1

   ! Step 1 in Section 5.2's procedure
   do i = 1,numR
      A(i,1:numP)=Rr(i,1:numP)           ! Initialize A=R
      Ainv(i,i)=1                        ! Initialize Ainv as an identity function (that is C in Step 1).
      if (i>1) then
         rownumber(i)=rownumber(i-1)+1   ! Initialize rownumber (that is d=1,2,...,K in Step 1)
      end if
   end do
   rowrank=numR                          ! Initialize rowrank=numR (that is M=K in Step 1).


   ! Step 2 in Section 5.2's procedure
   do i=1,numR
      do k=i+1,numR
         if (A(i,i).EQ.0 .and. A(k,i).NE.0) then   ! Step (i) in Step 2 (notations i and k are different from Step 2)
            do j=1,max(numR,numP)                  ! Swap rows in A and Ainv (C in Step 2)
               temp=A(i,j)
               A(i,j)=A(k,j)
               A(k,j)=temp
               temp=Ainv(i,j)
               Ainv(i,j)=Ainv(k,j)
               Ainv(k,j)=temp
            end do
            do j=1,(numP+1)                        ! Swap rows in Rr (this actually corresponds to Step 5)
               temp=Rr(i,j)
               Rr(i,j)=Rr(k,j)
               Rr(k,j)=temp
            end do      
               temp=rownumber(i)                   ! Swap elements in rownumber (d in Step 2)
               rownumber(i)=rownumber(k)
               rownumber(k)=temp       
         end if
      end do  

      if (A(i,i).NE.0) then                        ! Step (ii) in Step 2 (notation i is different from Step 2)
         temp=A(i,i)
         do j=1,max(numR,numP)
            A(i,j)=A(i,j)/temp
            Ainv(i,j)=Ainv(i,j)/temp
         end do

         do k=1,numR                               ! Step (iii) in Step 2 (notations i and k are different from Step 2)
            if (k.NE.i) then
               temp=A(k,i)
               do j=1,max(numR,numP)
                  A(k,j)=A(k,j)-A(i,j)*temp
                  Ainv(k,j)=Ainv(k,j)-Ainv(i,j)*temp
               end do
            end if
         end do
      end if

   end do


   ! Step 3 in Section 5.2's procedure
   do i=1,numR                                   
      if (sum(abs(A(i,1:numP))).EQ.0) then
         rowrank=rowrank-1
      end if
   end do

 
   ! Step 4 in Section 5.2's procedure
   do i=1,numR
      do k=i+1,numR
         if (sum(abs(A(i,1:numP))).EQ.0 .and. sum(abs(A(k,1:numP))).NE.0) then

            do j=1,max(numR,numP)                   ! Swap rows in A and Ainv (C in Step 4)             
               temp=A(i,j)
               A(i,j)=A(k,j)
               A(k,j)=temp
               temp=Ainv(i,j)
               Ainv(i,j)=Ainv(k,j)
               Ainv(k,j)=temp
            end do
            do j=1,(numP+1)                         ! Swap rows in Rr (this actually corresponds to Step 5)   
               temp=Rr(i,j)
               Rr(i,j)=Rr(k,j)
               Rr(k,j)=temp
            end do      
               temp=rownumber(i)                    ! Swap elements in rownumber (d in Step 4)
               rownumber(i)=rownumber(k)
               rownumber(k)=temp   
         end if   
      end do    
   end do
   
   do i=1,numR
      do j=1,numR
         transR(i,j)=Ainv(i,rownumber(j))           ! transR is C in Section 5.2
      end do    
      if (i>rowrank .and. sum(abs(transR(i,1:numR))).EQ.sum(transR(i,1:numR))) then    ! Check whether the hypothesis input in valid
         error=8
      end if
      transcon(i)=sum(Ainv(i,1:numR)*constant(1:numR))  ! This is for the constant used in the linear dependence of beta wave on beta bar.
      
   end do 
   
   
deallocate(A,Ainv,rownumber)   
end subroutine Mrank




! This is the function for constrained Gibbs sampler
subroutine gibbsampler(k,rowrank,bet,transcon,B,invbetacov,transR,f_or_c,Num)

   ! define variables
   implicit none
   integer                                          :: N, i, j, time
   integer, intent(in)                              :: k, rowrank
   integer, dimension(:), intent(out)               :: Num
   
   real                                             :: lower, upper, large, small, p, GG, temp
   real, dimension(:), intent(in)                   :: bet, transcon   
   real, allocatable, dimension(:)                  :: f_or_c, gibbsample, conmeans
   real, dimension(:,:), intent(in)                 :: B, invbetacov, transR
   
    
    allocate(gibbsample(rowrank),conmeans(rowrank))
    
    ! use method (41) 
    if (k.LE.rowrank) then     
    
       N=4000                 ! Initialize the sample size of the Gibbs sampler (Step 3 in Section 5.6)
  777  f_or_c(k)=0            ! This variable is used for both fit and complexity.
       gibbsample=0           ! initial value of the gibbs sampler (beta=0 in Step 3 in Section 5.6)

       do time=1,N+100    ! "time" (t) denotes the iteration index in the Gibbs sampler, 100 denotes the burn-in period.
          do i=1,rowrank  ! betas with a length of rowrank (M) need to be sampled (that is beta bar in Section 5)

          ! "conmeans" is the mean of conditional prior or posterior normal distribution, see equation (39)
             conmeans(i)=bet(i)+sum(B(i,1:rowrank)*(gibbsample(1:rowrank)-bet(1:rowrank)))-B(i,i)*(gibbsample(i)-bet(i))
                         
             if (i<k) then ! When computing the probability (41), beta_{i},i=1,...,k-1 have to be constrained (larger than 0) 

                ! lower bound is 0, but it has to be standardized for the standard normal function in Fortran.
                lower=(0-conmeans(i))/sqrt(1/invbetacov(i,i)) 
                GG=RNUNF() ! This function generates a pseudorandom number from a uniform (0,1) distribution
                
                ! "ANORDF" returns a probability that a normal random variable takes a value less than or equal to "lower"
                p=ANORDF(lower)+GG*(1-ANORDF(lower))
               
                ! Avoid p=1 and p=0, which are not accapted by "ANORIN" function; 
				! These are also not realistic situations. p=1 means the lower bound is infinity and p=0 means the upper bound is negative infinity.
				! But this can happen in the program because of the precision
                if (p>0.999999) then 
                   p=0.999999
                end if
                if (p<0.000001) then
                   p=0.000001
                end if
                
                ! "ANORIN" evaluates the inverse of the strandard normal distribution function.
                temp=ANORIN(p)
                ! Gibbs sample of beta_{k} is then obtained
                gibbsample(i)=temp*sqrt(1/invbetacov(i,i))+conmeans(i) 

             else ! if i>k, sample is unconstrained.
                GG=RNNOF()
		        gibbsample(i)=GG*sqrt(1/invbetacov(i,i))+conmeans(i)
             end if  

          end do   ! one iteration of gibbs sample ends. 

              
          ! Compute P(beta_{k}>0 |beta_{1}>0,...,beta_{k-1}>0) using (41)    
          if (time>100) then        ! discard first 100 iterations         
                temp=conmeans(k)/sqrt(1/invbetacov(k,k))
                ! ANORDF(temp) is the probability P(beta_{k}>0|beta_{i}^{t}>0,i<k), see equation (41) in the paper 
                f_or_c(k)=f_or_c(k)+ANORDF(temp)/real(N) ! Sum P(beta_{k}>0)/N to obtain the average of P(beta_{k}>0).
          end if  
       end do   
             

       ! Reset the number of iterations (sample size T2) based on the value of the estimation. See Table 2 in the paper
       if (f_or_c(k)<0.166 .AND. N==4000) then
       N=8000
       go to 777
       else if (f_or_c(k)<0.0417 .AND. N==8000) then
       N=12000
       go to 777
       else if (f_or_c(k)<0.00833 .AND. N==12000) then
       N=18000
       go to 777
       else if (f_or_c(k)<0.00139 .AND. N==18000) then
       N=25000
       go to 777
       else if (f_or_c(k)<0.000198 .AND. N==25000) then
       N=32000
       go to 777
       else if (f_or_c(k)<0.0000248 .AND. N==32000) then 
       N=100000
       go to 777
       end if
       Num(k)=N ! final sample size in "k" constraints
   	
	! use method (40) 
	else               
		   
       N=3000         ! Initialize the sample size of the Gibbs sampler (Step 3 in Section 5.6)
  888  f_or_c(k)=0    ! This variable is used both for fit and for complexity.
       gibbsample=0   ! initial value of the gibbs sampler (beta=0 in Step 3 in Section 5.6)

       do time=1,N+100 ! "time" (t) denotes the iteration index in the gibbs sampler, 100 denotes the burn-in period.
          do i=1,rowrank  ! numR betas need to be sampled (that is beta bar in Section 5)
          
          ! "conmeans" is the mean of the conditional prior or posterior normal distribution, see equation (39)
             conmeans(i)=bet(i)+sum(B(i,1:rowrank)*(gibbsample(1:rowrank)-bet(1:rowrank)))-B(i,i)*(gibbsample(i)-bet(i))
             
             large=0        ! Initialize the lower bound
             small=10**20   ! Initialize the upper bound   
                    
             if (k>(rowrank+1)) then   ! For each constraint k>M+1, define upper and lower bounds for beta bar.
                 do j=(rowrank+1),(k-1)
                    if (transR(j,i)<0) then
                       large=max(large,(transR(j,i)*gibbsample(i)-sum(transR(j,1:rowrank)*gibbsample)-transcon(j))/transR(j,i))
                    end if
                    if (transR(j,i)>0) then
                       small=min(small,(transR(j,i)*gibbsample(i)-sum(transR(j,1:rowrank)*gibbsample)-transcon(j))/transR(j,i))
                    end if
                 end do
             end if   
			  
			    ! standardize
                lower=(large-conmeans(i))/sqrt(1/invbetacov(i,i)) 
                upper=(small-conmeans(i))/sqrt(1/invbetacov(i,i)) 

                GG=RNUNF() ! This function generates a pseudorandom number from a uniform (0,1) distribution            
                ! "ANORDF" returns a probability that a normal random variable takes a value less than or equal to "lower"
                p=ANORDF(lower)+GG*(ANORDF(upper)-ANORDF(lower))
               
                ! Avoid p=1 and p=0, which are not accapted by "ANORIN" function; 
				! These are also not realistic situations. p=1 means the lower bound is infinity and p=0 means the upper bound is negative infinity.
				! But this can happen in the program because of the precision
                if (p>0.999999) then
                p=0.999999
                end if
                if (p<0.000001) then
                p=0.000001
                end if
                
                ! "ANORIN" evaluates the inverse of the strandard normal distribution function.
                temp=ANORIN(p)
                ! Gibbs sample of beta_{k} is then obtained
                gibbsample(i)=temp*sqrt(1/invbetacov(i,i))+conmeans(i) 
 
          end do   ! one iteration of gibbs sample ends. 
              
          ! Compute P(beta_{k}>0 |beta_{1}>0,...,beta_{k-1}>0) using (40)    
          if (time>100) then     ! discard first 100 sample 
                if (-sum(transR(k,1:rowrank)*gibbsample)>transcon(k)) then
                   f_or_c(k)=f_or_c(k)+1/real(N)
                end if
          end if
       end do    
       

	   ! Reset the number of iterations (sample size T1) based on the value of the estimation. See Table 2 in the paper
       if (f_or_c(k)<0.166 .AND. N==3000) then
       N=9600
       go to 888
       else if (f_or_c(k)<0.0417 .AND. N==9600) then
       N=120000
       go to 888
       else if (f_or_c(k)<0.00833 .AND. N==120000) then
       N=360000
       go to 888
       else if (f_or_c(k)<0.00139 .AND. N==360000) then 
       N=2520000
       go to 888
       else if (f_or_c(k)<0.000198 .AND. N==2520000) then  
       N=20160000
	   go to 888
       else if (f_or_c(k)<0.0000248 .AND. N==20160000) then 
       N=100000000
       go to 888
       end if
       Num(k)=N      
    end if
deallocate(gibbsample,conmeans)
end subroutine gibbsampler 

end program BFSEM


